@@ -8,7 +8,6 @@
 # Date:     26 May 2020
 # Copyright (c) 2020 Red Hat
 
-set -o errexit
 set -o nounset
 umask 077
 
@@ -21,13 +20,12 @@ IP6_FULL_PREFIX_SIZE=48 # Size of IP6_SUBNET
 IP4_SUBNET=10.11
 IP4_PREFIX_SIZE=24 # Size of assigned prefixes
 IP4_FULL_PREFIX_SIZE=16 # Size of IP4_SUBNET
-VLAN_IDS=(1 2)
 GENERATED_NAME_PREFIX="xdptest"
 ALL_TESTS=""
 VERBOSE_TESTS=${V:-0}
+NUM_NS=2
 
 NEEDED_TOOLS="capinfos ethtool ip ping sed tc tcpdump timeout nc tshark"
-MAX_NAMELEN=15
 
 if [ -f "$TEST_CONFIG" ]; then
     source "$TEST_CONFIG"
@@ -43,21 +41,11 @@ fi
 SKIPPED_TEST=249
 
 # Global state variables that will be set by options etc below
-GENERATE_NEW=0
-CLEANUP_FUNC=
 STATEDIR=
-STATEFILE=
 CMD=
 NS=
-LEGACY_IP=1
-USE_VLAN=0
-RUN_ON_INNER=0
+NS_NAMES=()
 
-# State variables that are written to and read from statefile
-STATEVARS=(IP6_PREFIX IP4_PREFIX
-           INSIDE_IP6 INSIDE_IP4 INSIDE_MAC
-           OUTSIDE_IP6 OUTSIDE_IP4 OUTSIDE_MAC
-           ENABLE_IPV4 ENABLE_VLAN)
 IP6_PREFIX=
 IP4_PREFIX=
 INSIDE_IP6=
@@ -66,8 +54,8 @@ INSIDE_MAC=
 OUTSIDE_IP6=
 OUTSIDE_IP4=
 OUTSIDE_MAC=
-ENABLE_IPV4=0
-ENABLE_VLAN=0
+ALL_INSIDE_IP6=()
+ALL_INSIDE_IP4=()
 
 is_trace_attach_supported()
 {
@@ -207,8 +195,6 @@ start_background_no_stderr()
 
 start_background_ns_devnull()
 {
-    get_nsname && ensure_nsname "$NS"
-
     local TMP_FILE="${STATEDIR}/tmp_proc_$$_$RANDOM"
     setsid ip netns exec "$NS" env TESTENV_NAME="$NS" "$SETUP_SCRIPT" bash -c "$*" 1>/dev/null 2>${TMP_FILE} &
     local PID=$!
@@ -255,95 +241,21 @@ check_prereq()
     if [ "$max_locked_mem" != "unlimited" ]; then
 	ulimit -l unlimited || die "Unable to set ulimit"
     fi
+
+    mount -t bpf bpf /sys/fs/bpf/ || die "Unable to mount bpffs"
 }
 
-get_nsname()
+gen_nsname()
 {
-    local GENERATE=${1:-1}
+    local nsname
 
-    if [ -z "$NS" ]; then
-        [ -f "$STATEDIR/current" ] && NS=$(< "$STATEDIR/current")
+    while
+        nsname=$(printf "%s-%04x" "$GENERATED_NAME_PREFIX" $RANDOM)
+        [ -e "$STATEDIR/${nsname}.ns" ]
+    do true; done
 
-        if [ "$GENERATE" -eq "1" ] && [ -z "$NS" ] || [ "$GENERATE_NEW" -eq "1" ]; then
-            NS="$GENERATED_NAME_PREFIX"
-            while [ -e "$STATEDIR/${NS}.state" ]; do
-                NS=$(printf "%s-%04x" "$GENERATED_NAME_PREFIX" $RANDOM)
-            done
-        fi
-    fi
-
-    if [ "${#NS}" -gt "$MAX_NAMELEN" ]; then
-        die "Environment name '$NS' is too long (max $MAX_NAMELEN)"
-    fi
-
-    STATEFILE="$STATEDIR/${NS}.state"
-}
-
-ensure_nsname()
-{
-    [ -z "$NS" ] && die "No environment selected; use --name to select one or 'setup' to create one"
-    [ -e "$STATEFILE" ] || die "Environment for $NS doesn't seem to exist"
-
-    echo "$NS" > "$STATEDIR/current"
-
-    read_statefile
-}
-
-get_num()
-{
-    local num=1
-    if [ -f "$STATEDIR/highest_num" ]; then
-        num=$(( 1 + $(< "$STATEDIR/highest_num" )))
-    fi
-
-    echo $num > "$STATEDIR/highest_num"
-    printf "%x" $num
-}
-
-write_statefile()
-{
-    [ -z "$STATEFILE" ] && return 1
-    echo > "$STATEFILE"
-    for var in "${STATEVARS[@]}"; do
-        echo "${var}='$(eval echo '$'$var)'" >> "$STATEFILE"
-    done
-}
-
-read_statefile()
-{
-    local value
-    for var in "${STATEVARS[@]}"; do
-        value=$(source "$STATEFILE"; eval echo '$'$var)
-        eval "$var=\"$value\""
-    done
-}
-
-cleanup_setup()
-{
-    echo "Error during setup, removing partially-configured environment '$NS'" >&2
-    set +o errexit
-    ip netns del "$NS" 2>/dev/null
-    ip link del dev "$NS" 2>/dev/null
-    rm -f "$STATEFILE"
-}
-
-cleanup_teardown()
-{
-    echo "Warning: Errors during teardown, partial environment may be left" >&2
-}
-
-
-cleanup()
-{
-    [ -n "$CLEANUP_FUNC" ] && $CLEANUP_FUNC
-
-    local statefiles=("$STATEDIR"/*.state)
-
-    if [ "${#statefiles[*]}" -eq 1 ] && [ ! -e "${statefiles[0]}" ]; then
-        rm -f "${STATEDIR}/highest_num" "${STATEDIR}/current" \
-           "${STATEDIR}/trace_attach_support"
-        rmdir "$STATEDIR"
-    fi
+    touch "$STATEDIR/${nsname}.ns"
+    echo $nsname
 }
 
 iface_macaddr()
@@ -359,129 +271,107 @@ set_sysctls()
     local nscmd=
 
     [ -n "$in_ns" ] && nscmd="ip netns exec $in_ns"
-    local sysctls=(accept_dad
-                   accept_ra
-                   mldv1_unsolicited_report_interval
-                   mldv2_unsolicited_report_interval)
+    local sysctls_off_v6=(accept_dad
+                       accept_ra
+                       mldv1_unsolicited_report_interval
+                       mldv2_unsolicited_report_interval)
+    local sysctls_on=(forwarding)
 
-    for s in ${sysctls[*]}; do
+    for s in ${sysctls_off_v6[*]}; do
         $nscmd sysctl -w net.ipv6.conf.$iface.${s}=0 >/dev/null
     done
+    for s in ${sysctls_on[*]}; do
+        $nscmd sysctl -w net.ipv6.conf.$iface.${s}=1 >/dev/null
+        $nscmd sysctl -w net.ipv6.conf.all.${s}=1 >/dev/null
+        $nscmd sysctl -w net.ipv4.conf.$iface.${s}=1 >/dev/null
+        $nscmd sysctl -w net.ipv4.conf.all.${s}=1 >/dev/null
+    done
 }
 
-get_vlan_prefix()
+init_ns()
 {
-    # Split the IPv6 prefix, and add the VLAN ID to the upper byte of the fourth
-    # element in the prefix. This will break if the global prefix config doesn't
-    # have exactly three elements in it.
-    local prefix="$1"
-    local vid="$2"
-    (IFS=:; set -- $prefix; printf "%s:%s:%s:%x::" "$1" "$2" "$3" $(($4 + $vid * 4096)))
-}
+    local nsname=$1
+    local num=$2
+    local peername="testl-ve-$num"
 
-setup()
-{
-    get_nsname 1
-
-    [ -e "$STATEFILE" ] && die "Environment for '$NS' already exists"
-
-    local NUM=$(get_num "$NS")
-    local PEERNAME="testl-ve-$NUM"
-    [ -z "$IP6_PREFIX" ] && IP6_PREFIX="${IP6_SUBNET}:${NUM}::"
-    [ -z "$IP4_PREFIX" ] && IP4_PREFIX="${IP4_SUBNET}.$((0x$NUM))."
+    IP6_PREFIX="${IP6_SUBNET}:${num}::"
+    IP4_PREFIX="${IP4_SUBNET}.$((0x$num))."
 
     INSIDE_IP6="${IP6_PREFIX}2"
     INSIDE_IP4="${IP4_PREFIX}2"
     OUTSIDE_IP6="${IP6_PREFIX}1"
     OUTSIDE_IP4="${IP4_PREFIX}1"
 
-    CLEANUP_FUNC=cleanup_setup
+    ip netns add "$nsname"
+    ip link add dev "$nsname" type veth peer name "$peername"
+    set_sysctls $nsname
 
-    if ! mount | grep -q /sys/fs/bpf; then
-        mount -t bpf bpf /sys/fs/bpf/
-    fi
+    ethtool -K "$nsname" rxvlan off txvlan off gro on
+    ethtool -K "$peername" rxvlan off txvlan off gro on
 
-    ip netns add "$NS"
-    ip link add dev "$NS" type veth peer name "$PEERNAME"
-    OUTSIDE_MAC=$(iface_macaddr "$NS")
-    INSIDE_MAC=$(iface_macaddr "$PEERNAME")
-    set_sysctls $NS
+    OUTSIDE_MAC=$(iface_macaddr "$nsname")
+    INSIDE_MAC=$(iface_macaddr "$peername")
+    ip link set dev "$peername" netns "$nsname"
+    ip link set dev "$nsname" up
+    ip addr add dev "$nsname" "${OUTSIDE_IP6}/${IP6_PREFIX_SIZE}"
 
-    ethtool -K "$NS" rxvlan off txvlan off
-    ethtool -K "$PEERNAME" rxvlan off txvlan off
-    ip link set dev "$PEERNAME" netns "$NS"
-    ip link set dev "$NS" up
-    ip addr add dev "$NS" "${OUTSIDE_IP6}/${IP6_PREFIX_SIZE}"
-
-    ip -n "$NS" link set dev "$PEERNAME" name veth0
-    ip -n "$NS" link set dev lo up
-    ip -n "$NS" link set dev veth0 up
-    set_sysctls veth0 "$NS"
-    ip -n "$NS" addr add dev veth0 "${INSIDE_IP6}/${IP6_PREFIX_SIZE}"
+    ip -n "$nsname" link set dev "$peername" name veth0
+    ip -n "$nsname" link set dev lo up
+    ip -n "$nsname" link set dev veth0 up
+    set_sysctls veth0 "$nsname"
+    ip -n "$nsname" addr add dev veth0 "${INSIDE_IP6}/${IP6_PREFIX_SIZE}"
 
     # Prevent neighbour queries on the link
-    ip neigh add "$INSIDE_IP6" lladdr "$INSIDE_MAC" dev "$NS" nud permanent
-    ip -n "$NS" neigh add "$OUTSIDE_IP6" lladdr "$OUTSIDE_MAC" dev veth0 nud permanent
+    ip neigh add "$INSIDE_IP6" lladdr "$INSIDE_MAC" dev "$nsname" nud permanent
+    ip -n "$nsname" neigh add "$OUTSIDE_IP6" lladdr "$OUTSIDE_MAC" dev veth0 nud permanent
 
-    # Add route for whole test subnet, to make it easier to communicate between
-    # namespaces
-    ip -n "$NS" route add "${IP6_SUBNET}::/$IP6_FULL_PREFIX_SIZE" via "$OUTSIDE_IP6" dev veth0
+    ip addr add dev "$nsname" "${OUTSIDE_IP4}/${IP4_PREFIX_SIZE}"
+    ip -n "$nsname" addr add dev veth0 "${INSIDE_IP4}/${IP4_PREFIX_SIZE}"
+    ip neigh add "$INSIDE_IP4" lladdr "$INSIDE_MAC" dev "$nsname" nud permanent
+    ip -n "$nsname" neigh add "$OUTSIDE_IP4" lladdr "$OUTSIDE_MAC" dev veth0 nud permanent
 
-    if [ "$LEGACY_IP" -eq "1" ]; then
-        ip addr add dev "$NS" "${OUTSIDE_IP4}/${IP4_PREFIX_SIZE}"
-        ip -n "$NS" addr add dev veth0 "${INSIDE_IP4}/${IP4_PREFIX_SIZE}"
-        ip neigh add "$INSIDE_IP4" lladdr "$INSIDE_MAC" dev "$NS" nud permanent
-        ip -n "$NS" neigh add "$OUTSIDE_IP4" lladdr "$OUTSIDE_MAC" dev veth0 nud permanent
-        ip -n "$NS" route add "${IP4_SUBNET}/${IP4_FULL_PREFIX_SIZE}" via "$OUTSIDE_IP4" dev veth0
-        ENABLE_IPV4=1
-    else
-        ENABLE_IPV4=0
-    fi
+    # Add default routes inside the ns
+    ip -n "$nsname" route add default via $OUTSIDE_IP4 dev veth0
+    ip -n "$nsname" -6 route add default via $OUTSIDE_IP6 dev veth0
 
-    if [ "$USE_VLAN" -eq "1" ]; then
-        ENABLE_VLAN=1
-        for vid in "${VLAN_IDS[@]}"; do
-            local vlpx="$(get_vlan_prefix "$IP6_PREFIX" "$vid")"
-            local inside_ip="${vlpx}2"
-            local outside_ip="${vlpx}1"
-            ip link add dev "${NS}.$vid" link "$NS" type vlan id "$vid"
-            ip link set dev "${NS}.$vid" up
-            ip addr add dev "${NS}.$vid" "${outside_ip}/${IP6_PREFIX_SIZE}"
-            ip neigh add "$inside_ip" lladdr "$INSIDE_MAC" dev "${NS}.$vid" nud permanent
-            set_sysctls "${NS}/$vid"
+    ALL_INSIDE_IP4+=($INSIDE_IP4)
+    ALL_INSIDE_IP6+=($INSIDE_IP6)
+}
 
-            ip -n "$NS" link add dev "veth0.$vid" link "veth0" type vlan id "$vid"
-            ip -n "$NS" link set dev "veth0.$vid" up
-            ip -n "$NS" addr add dev "veth0.$vid" "${inside_ip}/${IP6_PREFIX_SIZE}"
-            ip -n "$NS" neigh add "$outside_ip" lladdr "$OUTSIDE_MAC" dev "veth0.$vid" nud permanent
-            set_sysctls "veth0/$vid" "$NS"
-        done
-    else
-        ENABLE_VLAN=0
-    fi
+setup()
+{
+    local nsname
 
-    write_statefile
+    set -o errexit
 
-    CLEANUP_FUNC=
+    check_prereq
+
+    for i in $(seq $NUM_NS); do
+        nsname=$(gen_nsname)
+        init_ns $nsname $i
+        NS_NAMES+=($nsname)
+    done
+
+    set +o errexit
+
+    NS=$nsname
+}
+
+teardown_ns()
+{
+    local nsname=$1
+
+    ip link del dev "$nsname"
+    ip netns del "$nsname"
+    [ -d "/sys/fs/bpf/$nsname" ] && rmdir "/sys/fs/bpf/$nsname" || true
 
-    echo "$NS" > "$STATEDIR/current"
 }
 
 teardown()
 {
-    get_nsname && ensure_nsname "$NS"
-
-    CLEANUP_FUNC=cleanup_teardown
-
-    ip link del dev "$NS"
-    ip netns del "$NS"
-    rm -f "$STATEFILE"
-    [ -d "/sys/fs/bpf/$NS" ] && rmdir "/sys/fs/bpf/$NS" || true
-
-    if [ -f "$STATEDIR/current" ]; then
-        local CUR=$(< "$STATEDIR/current" )
-        [[ "$CUR" == "$NS" ]] && rm -f "$STATEDIR/current"
-    fi
+    for ns in "${NS_NAMES[@]}"; do
+        teardown_ns $ns
+    done
 
     for f in ${STATEDIR}/proc/*; do
         if [ -f "$f" ]; then
@@ -491,14 +381,10 @@ teardown()
     done
 
     rm -rf "$STATEDIR"
-
-    CLEANUP_FUNC=
 }
 
 ns_exec()
 {
-    get_nsname && ensure_nsname "$NS"
-
     ip netns exec "$NS" env TESTENV_NAME="$NS" "$SETUP_SCRIPT" "$@"
 }
 
@@ -554,7 +440,6 @@ run_tests()
     local TESTS="$*"
     local ret=0
     [ -z "$TESTS" ] && TESTS="$ALL_TESTS"
-    setup || return 1
 
     echo "    Running tests from $TEST_DEFINITIONS"
 
@@ -576,14 +461,22 @@ usage()
 
 if [ "$EUID" -ne "0" ]; then
     if command -v sudo >/dev/null 2>&1; then
-        exec sudo env V=${VERBOSE_TESTS} "$0" "$@"
+        exec sudo env V=${VERBOSE_TESTS} DEBUG_TESTENV=${DEBUG_TESTENV:-0} "$0" "$@"
     else
         die "Tests should be run as root"
     fi
+else
+    if [ "${DID_UNSHARE:-0}" -ne "1" ]; then
+        echo "    Executing tests in separate net- and mount namespaces" >&2
+        exec env DID_UNSHARE=1 unshare -n -m "$0" "$@"
+    fi
 fi
 
 export XDP_FILTER
 export XDP_LOADER
+export XDP_BENCH
+export XDP_MONITOR
+export XDP_TRAFFICGEN
 export XDPDUMP
 
 TEST_DEFINITIONS="${1:-}"
@@ -594,5 +487,11 @@ TOOL_TESTS_DIR="$(dirname "$TEST_DEFINITIONS")"
 
 shift
 trap teardown EXIT
-check_prereq
+setup
+
+if [ "${DEBUG_TESTENV:-0}" -eq "1" ] && [ -n "$SHELL" ]; then
+    echo "Entering interactive testenv debug - Ctrl-D to exit and resume test execution"
+    $SHELL
+fi
+
 run_tests "$@"
